/*
 * Copyright (c) 2012 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is confidential property of Nordic Semiconductor. The use,
 * copying, transfer or disclosure of such information is prohibited except by express written
 * agreement with Nordic Semiconductor.
 *
 */

/** @example Board/nrf6310/s120/experimental/ble_app_hrs_c/main.c
 *
 * @brief BLE Heart Rate Collector application main file.
 *
 * This file contains the source code for a sample heart rate collector.
 */

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf_sdm.h"
#include "ble.h"
#include "ble_db_discovery.h"
#include "softdevice_handler.h"
#include "app_util.h"
#include "app_error.h"
#include "ble_advdata_parser.h"
#include "boards.h"
#include "nrf_gpio.h"
#include "pstorage.h"
#include "device_manager.h"
#include "debug.h"
#include "ble_hrs_c.h"
#include "ble_bas_c.h"
#include "app_util.h"
#include "simple_uart.h"

#define LCD_LLEN 						16

//#define BUTTON_1						1
//#define BUTTON_PULL						2
//
//#define BOND_DELETE_ALL_BUTTON_ID        BUTTON_1                                       /**< Button used for deleting all bonded centrals during startup. */

#define SCAN_LED_PIN_NO                 LED_RGB_RED                     /**< Is on when device is scanning. */
#define CONNECTED_LED_PIN_NO            LED_RGB_GREEN                   /**< Is on when device has connected. */
#define ASSERT_LED_PIN_NO               LED_RGB_BLUE					/**< Is on when application has asserted. */

#define APPL_LOG                         debug_log                                      /**< Debug logger macro that will be used in this file to do logging of debug information over UART. */

#define SEC_PARAM_BOND                   1                                              /**< Perform bonding. */
#define SEC_PARAM_MITM                   1                                              /**< Man In The Middle protection not required. */
#define SEC_PARAM_IO_CAPABILITIES        BLE_GAP_IO_CAPS_NONE                           /**< No I/O capabilities. */
#define SEC_PARAM_OOB                    0                                              /**< Out Of Band data not available. */
#define SEC_PARAM_MIN_KEY_SIZE           7                                              /**< Minimum encryption key size. */
#define SEC_PARAM_MAX_KEY_SIZE           16                                             /**< Maximum encryption key size. */

#define SCAN_INTERVAL                    0x00A0                                         /**< Determines scan interval in units of 0.625 millisecond. */
#define SCAN_WINDOW                      0x0050                                         /**< Determines scan window in units of 0.625 millisecond. */

#define MIN_CONNECTION_INTERVAL     MSEC_TO_UNITS(7.5, UNIT_1_25_MS)                    /**< Determines maximum connection interval in millisecond. */
#define MAX_CONNECTION_INTERVAL     MSEC_TO_UNITS(30, UNIT_1_25_MS)                     /**< Determines maximum connection interval in millisecond. */
#define SLAVE_LATENCY               0                                                   /**< Determines slave latency in counts of connection events. */
#define SUPERVISION_TIMEOUT         MSEC_TO_UNITS(4000, UNIT_10_MS)                     /**< Determines supervision time-out in units of 10 millisecond. */

#define TARGET_UUID                 0x180D                                              /**< Target device name that application is looking for. */
#define MAX_PEER_COUNT              DEVICE_MANAGER_MAX_CONNECTIONS                      /**< Maximum number of peer's application intends to manage. */
#define UUID16_SIZE                 2                                                   /**< Size of 16 bit UUID */

/**@breif Macro to unpack 16bit unsigned UUID from octet stream. */
#define UUID16_EXTRACT(DST,SRC)                                                                  \
        do                                                                                       \
        {                                                                                        \
            (*(DST)) = (SRC)[1];                                                                 \
            (*(DST)) <<= 8;                                                                      \
            (*(DST)) |= (SRC)[0];                                                                \
        } while(0)

/**@brief Variable length data encapsulation in terms of length and pointer to data */
typedef struct
{
    uint8_t               * p_data;          /**< Pointer to data. */
    uint16_t                data_len;        /**< Length of data. */
}data_t;

static ble_gap_scan_params_t   	m_scan_param;             /**< Scan parameters requested for scanning and connection. */
static uint16_t 				conn_hdl = 	BLE_CONN_HANDLE_INVALID;


/**
 * @brief Connection parameters requested for connection.
 */
static const ble_gap_conn_params_t m_connection_param =
{
    (uint16_t)MIN_CONNECTION_INTERVAL,   // Minimum connection
    (uint16_t)MAX_CONNECTION_INTERVAL,   // Maximum connection
    0,                                   // Slave latency
    (uint16_t)SUPERVISION_TIMEOUT        // Supervision time-out
};

static void scan_start(void);

#define APPL_LOG                        debug_log                 /**< Debug logger macro that will be used in this file to do logging of debug information over UART. */

// WARNING: The following macro MUST be un-defined (by commenting out the definition) if the user
// does not have a nRF6350 Display unit. If this is not done, the application will not work.
//#define APPL_LCD_PRINT_ENABLE                                     /**< In case you do not have a functional display unit, disable this flag and observe trace on UART. */

#ifdef APPL_LCD_PRINT_ENABLE

#define APPL_LCD_CLEAR                  nrf6350_lcd_clear         /**< Macro to clear the LCD display.*/
#define APPL_LCD_WRITE                  nrf6350_lcd_write_string  /**< Macro to write a string to the LCD display.*/

#else // APPL_LCD_PRINT_ENABLE

#define APPL_LCD_WRITE(...)             true                      /**< Macro to clear the LCD display defined to do nothing when @ref APPL_LCD_PRINT_ENABLE is not defined.*/
#define APPL_LCD_CLEAR(...)             true                      /**< Macro to write a string to the LCD display defined to do nothing when @ref APPL_LCD_PRINT_ENABLE is not defined.*/

#endif // APPL_LCD_PRINT_ENABLE


/**@brief Function for error handling, which is called when an error has occurred.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of error.
 *
 * @param[in] error_code  Error code supplied to the handler.
 * @param[in] line_num    Line number where the handler is called.
 * @param[in] p_file_name Pointer to the file name.
 */
void app_error_handler(uint32_t error_code, uint32_t line_num, const uint8_t * p_file_name)
{
    APPL_LOG("[APPL]: ASSERT: %s, %d, error 0x%08x\r\n", p_file_name, line_num, error_code);
    nrf_gpio_pin_clear(ASSERT_LED_PIN_NO);
    nrf_gpio_pin_set(SCAN_LED_PIN_NO);
    nrf_gpio_pin_set(CONNECTED_LED_PIN_NO);

    // This call can be used for debug purposes during development of an application.
    // @note CAUTION: Activating this code will write the stack to flash on an error.
    //                This function should NOT be used in a final product.
    //                It is intended STRICTLY for development/debugging purposes.
    //                The flash write will happen EVEN if the radio is active, thus interrupting
    //                any communication.
    //                Use with care. Un-comment the line below to use.
    // ble_debug_assert_handler(error_code, line_num, p_file_name);

    for(;;);
    // On assert, the system can only recover with a reset.
    NVIC_SystemReset();
}


/**@brief Function for asserts in the SoftDevice.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num     Line number of the failing ASSERT call.
 * @param[in] p_file_name  File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(0xDEADBEEF, line_num, p_file_name);
}

/**
 * @brief Parses advertisement data, providing length and location of the field in case
 *        matching data is found.
 *
 * @param[in]  Type of data to be looked for in advertisement data.
 * @param[in]  Advertisement report length and pointer to report.
 * @param[out] If data type requested is found in the data report, type data length and
 *             pointer to data will be populated here.
 *
 * @retval NRF_SUCCESS if the data type is found in the report.
 * @retval NRF_ERROR_NOT_FOUND if the data type could not be found.
 */
static uint32_t adv_report_parse(uint8_t type, data_t * p_advdata, data_t * p_typedata)
{
    uint32_t index = 0;
    uint8_t * p_data;

    p_data = p_advdata->p_data;

    while (index < p_advdata->data_len)
    {
        uint8_t field_length = p_data[index];
        uint8_t field_type = p_data[index+1];

        if (field_type == type)
        {
            p_typedata->p_data = &p_data[index+2];
            p_typedata->data_len = field_length-1;
            return NRF_SUCCESS;
        }
        index += field_length+1;
    }
    return NRF_ERROR_NOT_FOUND;
}


/**@brief Function for handling the Application's BLE Stack events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void on_ble_evt(ble_evt_t * p_ble_evt)
{
    uint32_t                err_code;
    const ble_gap_evt_t   * p_gap_evt = &p_ble_evt->evt.gap_evt;
    const ble_gattc_evt_t * p_gattc_evt = &p_ble_evt->evt.gattc_evt;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_ADV_REPORT:
        {
            data_t adv_data;
            data_t type_data;

            // Initialize advertisement report for parsing.
            adv_data.p_data = (uint8_t *)p_gap_evt->params.adv_report.data;
            adv_data.data_len = p_gap_evt->params.adv_report.dlen;

            uint8_t * adrs_ptr = &(p_gap_evt->params.adv_report.peer_addr.addr);

            APPL_LOG("[%02X %02X %02X %02X %02X %02X]\r\n",
                                adrs_ptr[0], adrs_ptr[1], adrs_ptr[2],
                                adrs_ptr[3], adrs_ptr[4], adrs_ptr[5]);


            const uint8_t temp[6]= {0xF9, 0x4D, 0x1A, 0xDE, 0x0D, 0xED};
            if(strncmp(adrs_ptr,temp,6)==0){
            	// Stop scanning.
				err_code = sd_ble_gap_scan_stop();
				if (err_code != NRF_SUCCESS)
				{
					APPL_LOG("[APPL]: Scan stop failed, reason %d\r\n", err_code);
				}
				nrf_gpio_pin_set(SCAN_LED_PIN_NO);

				m_scan_param.selective = 0;

				// Initiate connection.
				err_code = sd_ble_gap_connect(&p_gap_evt->params.adv_report.\
											   peer_addr,
											   &m_scan_param,
											   &m_connection_param);

				if (err_code != NRF_SUCCESS)
				{
					APPL_LOG("[APPL]: Connection Request Failed, reason %d\r\n", err_code);
				}
				break;

            }

            break;
        }
        case BLE_GAP_EVT_TIMEOUT:
            if (p_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN)
            {
                APPL_LOG("[APPL]: Connection Request Timedout.\r\n");
            }
            break;
        case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
            // Accepting parameters requested by peer.
            err_code = sd_ble_gap_conn_param_update(p_gap_evt->conn_handle,
                                                    &p_gap_evt->params.conn_param_update_request.conn_params);
            APP_ERROR_CHECK(err_code);
            break;
        case BLE_GAP_EVT_CONNECTED:
        	//Primary Service discovery
        	printf("Connected\n");
        	conn_hdl = p_gap_evt->conn_handle;
        	err_code = sd_ble_gattc_primary_services_discover(p_gap_evt->conn_handle, 0x0001, NULL );
        	APP_ERROR_CHECK(err_code);
        	break;
        case BLE_GAP_EVT_DISCONNECTED:
        	conn_hdl = BLE_CONN_HANDLE_INVALID;
        	break;
        case BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP:{
        	uint32_t end_hnd;
        	printf("Count:%d ",p_gattc_evt->params.prim_srvc_disc_rsp.count);
        	for(uint32_t i =0; i < p_gattc_evt->params.prim_srvc_disc_rsp.count; i++){
				printf("Service:%04X ",p_gattc_evt->params.prim_srvc_disc_rsp.services[i].uuid.uuid);
				printf("Start:%d ",p_gattc_evt->params.prim_srvc_disc_rsp.services[i].handle_range.start_handle);
				end_hnd = p_gattc_evt->params.prim_srvc_disc_rsp.services[i].handle_range.end_handle;
				printf("End:%d ",end_hnd);
        	}
        	printf("\n");
        	if(end_hnd != 0xFFFF){
        		err_code = sd_ble_gattc_primary_services_discover(p_gattc_evt->conn_handle, end_hnd, NULL );
				APP_ERROR_CHECK(err_code);
        	}else{
        		err_code = sd_ble_gattc_read(p_gap_evt->conn_handle, 0x0011, 0);
				APP_ERROR_CHECK(err_code);
        	}
        	break;}
        case BLE_GATTC_EVT_READ_RSP:
        	printf("ReadHdl:%d ",p_gattc_evt->params.read_rsp.handle + p_gattc_evt->params.read_rsp.offset);
        	for(uint32_t i = 0; i< p_gattc_evt->params.read_rsp.len; i++){
				printf("ReadData:%02X", p_gattc_evt->params.read_rsp.data[i]);
			}
        	printf("\n");
        	break;
        case BLE_GATTC_EVT_HVX:
        	printf("HvxHdl:%d Type:%d ",p_gattc_evt->params.hvx.handle, p_gattc_evt->params.hvx.type);
        	for(uint32_t i = 0; i< p_gattc_evt->params.hvx.len; i++){
        		printf("HvxData:%02X", p_gattc_evt->params.hvx.data[i]);
        	}
        	printf("\n");
        	break;
        case BLE_GATTC_EVT_WRITE_RSP:
        	printf("WrHdl:%d ",p_gattc_evt->params.write_rsp.handle);
			for(uint32_t i = 0; i< p_gattc_evt->params.write_rsp.len; i++){
				printf("HvxData:%02X", p_gattc_evt->params.write_rsp.data[i]);
			}
			printf("\n");
        	break;
        default:
            break;
    }
}


/**@brief Function for dispatching a BLE stack event to all modules with a BLE stack event handler.
 *
 * @details This function is called from the scheduler in the main loop after a BLE stack event has
 *  been received.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void ble_evt_dispatch(ble_evt_t * p_ble_evt)
{

    on_ble_evt(p_ble_evt);
}


/**@brief Function for dispatching a system event to interested modules.
 *
 * @details This function is called from the System event interrupt handler after a system
 *          event has been received.
 *
 * @param[in]   sys_evt   System stack event.
 */
static void sys_evt_dispatch(uint32_t sys_evt)
{
    pstorage_sys_event_handler(sys_evt);
}


/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void)
{
    uint32_t err_code;

    // Initialize the SoftDevice handler module.
    SOFTDEVICE_HANDLER_INIT(NRF_CLOCK_LFCLKSRC_XTAL_20_PPM, false);

    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);

    // Register with the SoftDevice handler module for System events.
    err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);
}

/**@brief Function for the LEDs initialization.
 *
 * @details Initializes all LEDs used by this application.
 */
static void leds_init(void)
{
    nrf_gpio_cfg_output(SCAN_LED_PIN_NO);
    nrf_gpio_cfg_output(CONNECTED_LED_PIN_NO);
    nrf_gpio_cfg_output(ASSERT_LED_PIN_NO);
    nrf_gpio_pin_set(SCAN_LED_PIN_NO);
    nrf_gpio_pin_set(CONNECTED_LED_PIN_NO);
    nrf_gpio_pin_set(ASSERT_LED_PIN_NO);
}


/** @brief Function for the Power manager.
 */
static void power_manage(void)
{
    uint32_t err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);
}

/**@breif Function to start scanning.
 */
static void scan_start(void)
{
	uint32_t err_code;

	// No devices in whitelist, hence non selective performed.
	m_scan_param.active       = 0;            // Active scanning set.
	m_scan_param.selective    = 0;            // Selective scanning not set.
	m_scan_param.interval     = SCAN_INTERVAL;// Scan interval.
	m_scan_param.window       = SCAN_WINDOW;  // Scan window.
	m_scan_param.p_whitelist  = NULL;         // No whitelist provided.
	m_scan_param.timeout      = 0x0000;       // No timeout.

    err_code = sd_ble_gap_scan_start(&m_scan_param);
    APP_ERROR_CHECK(err_code);
    
    bool lcd_write_status = APPL_LCD_WRITE("Scanning", 8, LCD_UPPER_LINE, 0);
    if (!lcd_write_status)
    {
        APPL_LOG("[APPL]: LCD Write failed!\r\n");
    }

    nrf_gpio_pin_clear(SCAN_LED_PIN_NO);
}

void uart_handler(uint8_t rx_char){
	uint32_t err_code;
	simple_uart_put(rx_char);
	switch(rx_char){
		case 'a':
			if(conn_hdl != BLE_CONN_HANDLE_INVALID){
				ble_gattc_write_params_t wr_param;
				uint8_t notif_enb[2] = {0x01,0x00};

				wr_param.write_op = BLE_GATT_OP_WRITE_REQ;
				wr_param.flags = BLE_GATT_EXEC_WRITE_FLAG_PREPARED_WRITE;
				wr_param.handle = 0x12;
				wr_param.offset = 0;
				wr_param.len = 2;
				wr_param.p_value = notif_enb;

				sd_ble_gattc_write(conn_hdl, &wr_param);
				APP_ERROR_CHECK(err_code);
			}
		break;
		case 'b':
			if(conn_hdl != BLE_CONN_HANDLE_INVALID){
				ble_gattc_write_params_t wr_param;
				uint8_t alrm_wr[4] = {0x09,0x00,0x00,0x00};

				wr_param.write_op = BLE_GATT_OP_WRITE_REQ;
				wr_param.flags = BLE_GATT_EXEC_WRITE_FLAG_PREPARED_WRITE;
				wr_param.handle = 0x11;
				wr_param.offset = 0;
				wr_param.len = 4;
				wr_param.p_value = alrm_wr;

				err_code = sd_ble_gattc_write(conn_hdl, &wr_param);
				APP_ERROR_CHECK(err_code);
			}
		break;
		case 'c':
			if(conn_hdl != BLE_CONN_HANDLE_INVALID){
				err_code = sd_ble_gattc_read(conn_hdl, 0x0011, 0);
				APP_ERROR_CHECK(err_code);
			}
			break;
		default: break;
	}
}

int main(void)
{
    // Initialization of various modules.
    debug_init();
    leds_init();
    uart_handler_init(&uart_handler);
    ble_stack_init();

    printf("Start\n");

    // Start scanning for peripherals and initiate connection
    scan_start();

    for (;;)
    {
        power_manage();
    }
}

/*
**PRIMARY**
attr handle: 0x0001, end grp handle: 0x0007 uuid: 00001800-0000-1000-8000-00805f9b34fb
attr handle: 0x0008, end grp handle: 0x000b uuid: 00001801-0000-1000-8000-00805f9b34fb
attr handle: 0x000c, end grp handle: 0xffff uuid: 00001523-1212-efde-1523-785feabcd123
**CHARACTERISTICS**
handle: 0x0002, char properties: 0x0a, char value handle: 0x0003, uuid: 00002a00-0000-1000-8000-00805f9b34fb
handle: 0x0004, char properties: 0x02, char value handle: 0x0005, uuid: 00002a01-0000-1000-8000-00805f9b34fb
handle: 0x0006, char properties: 0x02, char value handle: 0x0007, uuid: 00002a04-0000-1000-8000-00805f9b34fb
handle: 0x0009, char properties: 0x20, char value handle: 0x000a, uuid: 00002a05-0000-1000-8000-00805f9b34fb
handle: 0x000d, char properties: 0x12, char value handle: 0x000e, uuid: 00001524-1212-efde-1523-785feabcd123
handle: 0x0010, char properties: 0x1a, char value handle: 0x0011, uuid: 00001525-1212-efde-1523-785feabcd123
**HANDLES**
handle: 0x0001, uuid: 2800							GATT Primary Service Declaration
handle: 0x0002, uuid: 2803							GATT Characteristic Declaration
handle: 0x0003, uuid: 2a00							Device Name
handle: 0x0004, uuid: 2803							GATT Characteristic Declaration
handle: 0x0005, uuid: 2a01							Appearance
handle: 0x0006, uuid: 2803							GATT Characteristic Declaration
handle: 0x0007, uuid: 2a04							Peripheral Preferred Connection Parameters
handle: 0x0008, uuid: 2800							GATT Primary Service Declaration
handle: 0x0009, uuid: 2803							GATT Characteristic Declaration
handle: 0x000a, uuid: 2a05							Service Changed
handle: 0x000b, uuid: 2902							Client Characteristic Configuration (Indication and/or Notification dis/enabling)
handle: 0x000c, uuid: 2800							GATT Primary Service Declaration
handle: 0x000d, uuid: 2803							GATT Characteristic Declaration
handle: 0x000e, uuid: 00001524-1212-efde-1523-785feabcd123
handle: 0x000f, uuid: 2902							Client Characteristic Configuration (Indication and/or Notification dis/enabling)
handle: 0x0010, uuid: 2803							GATT Characteristic Declaration
handle: 0x0011, uuid: 00001525-1212-efde-1523-785feabcd123
handle: 0x0012, uuid: 2902							Client Characteristic Configuration (Indication and/or Notification dis/enabling)
*********
 */
